﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Rocket : MonoBehaviour
{
    [SerializeField]
    private float _boostSpeed;
    [SerializeField]
    private float _rotationTime;
    [SerializeField]
    private AnimationCurve _rotationCurve;
    [SerializeField]
    private float _boostTime;

    [SerializeField]
    private Rigidbody2D _rigidbody;

    [SerializeField]
    private Collider2D _collider;

    [SerializeField]
    private Planet _startPlanet;

    [SerializeField] private Sprite[] _powSprites;
    [SerializeField] private SpriteRenderer _powRenderer;
    [SerializeField] private Animator _powAnimator;

    [SerializeField]
    private ParticleSystem _fireParticles;


    private Queue<float> _boostQueue = new Queue<float>();
    private Queue<float> _rotateQueue = new Queue<float>();

    private IEnumerator _boostRoutine;
    private IEnumerator _rotateRoutine;

    private List<DelayedAction> _delayedActions = new List<DelayedAction>();

    public float BoostSpeed
    {
        get { return _boostSpeed; }
    }

    public List<DelayedAction> DelayedActions
    {
        get { return _delayedActions; }
    }

    public float MaxHealth
    {
        get
        {
            return 100f;
        }
    }

    public float CurrentHealth
    {
        get;
        private set;
    }

    public Rigidbody2D Rigidbody
    {
        get { return _rigidbody; }
    }

    public float DelayTime
    {
        get
        {
            return Vector3.Distance(transform.position, _startPlanet.transform.position) / 40f;
        }
    }

    private void Start()
    {
        CurrentHealth = MaxHealth;
    }

    private void Update()
    {
        if (GameManager.Instance.GameHasFinished || CurrentHealth <= 0) return;

        if (GameSettings.GameMode == GameMode.Sandbox)
        {
            if (Input.GetKeyDown(KeyCode.Y))
            {
                AddBoost(2f);
            }
            if (Input.GetKeyDown(KeyCode.U))
            {
                AddRotation(45f);
            }
            if (Input.GetKeyDown(KeyCode.I))
            {
                AddRotation(0f);
            }
        }

        for(int i = _delayedActions.Count - 1; i >= 0; i--)
        {
            DelayedAction delayedAction = _delayedActions[i];
            delayedAction.Delay -= Time.deltaTime;
            if(delayedAction.Delay <= 0)
            {
                delayedAction.Action();
                _delayedActions.Remove(delayedAction);
            }
        }
    }

    private void OnCollisionEnter2D(Collision2D collision)
    {
        if (collision.otherCollider == _collider)
        {
            if (GameSettings.GameMode != GameMode.Sandbox)
            {
                CurrentHealth -= collision.relativeVelocity.magnitude * 5;
                if (CurrentHealth <= 0)
                {
                    StopAllCoroutines();
                    GameManager.Instance.RocketHasDied(this);
                }
            }

            int r = UnityEngine.Random.Range(0, _powSprites.Length);

            ContactPoint2D[] contacts = collision.contacts;

            if (contacts != null && contacts.Length > 0)
            {
                GameObject clone = Instantiate(_powAnimator.gameObject);
                SpriteRenderer spriteRenderer = clone.GetComponent<SpriteRenderer>();
                Animator animator = clone.GetComponent<Animator>();

                clone.transform.position = contacts[0].point;
                spriteRenderer.sprite = _powSprites[r];
                animator.SetBool("isPlaying", true);
            }

            if (AudioManager.IsAvailable) AudioManager.Instance.PlayCrash();
        }
    }

    public void GameFinished()
    {
        StopAllCoroutines();

    }

    public void AddDelayedBoost(float boost)
    {
        if (GameManager.Instance.GameHasFinished || CurrentHealth <= 0) return;

        _delayedActions.Add(new DelayedAction(() => AddBoost(boost), DelayTime, boost + "m/s", "Boost"));
    }

    public void AddDelayedRotation(float rotation)
    {
        if (GameManager.Instance.GameHasFinished || CurrentHealth <= 0) return;

        _delayedActions.Add(new DelayedAction(() => AddRotation(rotation), DelayTime, rotation + "°", "Rotation"));
    }

    public void AddBoost(float boost)
    {
        if (GameManager.Instance.GameHasFinished || CurrentHealth <= 0) return;

        if (_boostRoutine == null)
        {
            StartCoroutine(_boostRoutine = BoostRoutine(boost));
        }
        else
        {
            _boostQueue.Enqueue(boost);
        }
    }

    public void AddRotation(float rotation)
    {
        if (GameManager.Instance.GameHasFinished || CurrentHealth <= 0) return;

        if(_rotateRoutine == null)
        {
            StartCoroutine(_rotateRoutine = RotateRoutine(rotation));
        }
        else
        {
            _rotateQueue.Enqueue(rotation);
        }
    }

    private IEnumerator BoostRoutine(float boost)
    {
        if (AudioManager.IsAvailable) AudioManager.Instance.PlayRocketBoost();

        float boostPerSeconds = boost / _boostTime;

        while (boost >= 0)
        {
            _fireParticles.Emit(10);
            float boostThisFrame = boostPerSeconds * Time.deltaTime;
            boost -= boostThisFrame;
            _rigidbody.AddForce((transform.up * boostThisFrame * 100));
            yield return new WaitForFixedUpdate();
        }

        _boostRoutine = null;

        if (_boostQueue.Count > 0)
        {
            StartCoroutine(_boostRoutine = BoostRoutine(_boostQueue.Dequeue()));
        }
    }

    private IEnumerator RotateRoutine(float rotation)
    {
        if (AudioManager.IsAvailable) AudioManager.Instance.PlaySwipe();

        float startRotation = transform.rotation.eulerAngles.z;
        float targetRotation = transform.rotation.eulerAngles.z - rotation;
        float currentRotationTime = _rotationTime;

        //Reset current rocket rotation (to prevent drifting afterwards)
        _rigidbody.angularVelocity = 0f;

        while (currentRotationTime > 0f)
        {
            Vector3 rot = transform.rotation.eulerAngles;
            rot.z = Mathf.Lerp(startRotation, targetRotation, _rotationCurve.Evaluate(1 - currentRotationTime / _rotationTime));
            transform.rotation = Quaternion.Euler(rot);
            currentRotationTime -= Time.deltaTime;
            yield return new WaitForFixedUpdate();
        }

        _rotateRoutine = null;

        if(_rotateQueue.Count > 0)
        {
            StartCoroutine(_rotateRoutine = RotateRoutine(_rotateQueue.Dequeue()));
        }
    }

    public class DelayedAction
    {
        public Action Action;
        public float Delay;
        public string ReadableValue;
        public string Name;

        public DelayedAction(Action action, float delay, string readableValue, string name)
        {
            Action = action;
            Delay = delay;
            ReadableValue = readableValue;
            Name = name;
        }
    }
}
